/* Copyright 2012 Cloudseal Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudseal.client.tomcat;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletResponse;

import com.cloudseal.client.saml2.*;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.Session;
import org.apache.catalina.authenticator.FormAuthenticator;
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.deploy.LoginConfig;
import org.apache.catalina.realm.GenericPrincipal;
import org.apache.commons.io.IOUtils;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class CloudsealAuthenticator extends FormAuthenticator {

    public static final String AUTH_METHOD = "CLOUDSEAL";
    private static final String BROWSER_INITIATED_SSO_URL = "/logout.jsonp";
    private static final String ID = "com.cloudseal.client.tomcat.Authenticator.Id";
    private static final String AUDIENCE = "com.cloudseal.client.tomcat.Authenticator.Audience";

    private static final Log LOG = LogFactory.getLog(CloudsealAuthenticator.class);

    private IdentifierGenerator idGenerator;
    private AuthRequestBuilder authRequestBuilder;
    private AuthResponseValidator responseValidator;
    private IdpXmlParser idpXmlParser;

    private String idp;
    private String keystore;
    private String keystorePassword;
    private String key;
    private String keyPassword;
    private String appId;

    public CloudsealAuthenticator() throws Exception {
        setAlwaysUseSession(true);
    }

    @Override
    protected String getAuthMethod() {
        return AUTH_METHOD;
    }

    @Override
    protected void initInternal() throws LifecycleException {
        super.initInternal();
        synchronized (this) {
            LOG.info("Initializing");
            checkConfigParam(idp, "idp");
            checkConfigParam(keystore, "keystore");
            checkConfigParam(keystorePassword, "keystorePassword");
            checkConfigParam(key, "key");
            checkConfigParam(keyPassword, "keyPassword");
            try {
                AuthRequestContext authRequestContext = new AuthRequestContext();
                authRequestContext.setKeyName(key);
                authRequestContext.setKeyPassword(keyPassword);
                authRequestContext.setKeystore(getResource(keystore));
                authRequestContext.setKeystorePassword(keystorePassword);
                authRequestContext.setProviderName(appId);
                authRequestBuilder = new AuthRequestBuilder();
                authRequestBuilder.init(authRequestContext);
                responseValidator = new AuthResponseValidator();
                idpXmlParser = new IdpXmlParser();
                idpXmlParser.parse(getResource(idp));
                idGenerator = new IdentifierGenerator();
            } catch (Exception e) {
                throw new LifecycleException("Failed to initialize", e);
            }
        }
    }

    private void checkConfigParam(String param, String paramName) throws LifecycleException {
        if (param == null || param.length() < 1) {
            throw new LifecycleException("Missing configuration parameter: " + paramName);
        }
    }

    @Override
    public void invoke(Request request, Response response) throws IOException, ServletException {
        LOG.info("invoke()");

        // single log out?
        if (request.getRequestURL().toString().endsWith(BROWSER_INITIATED_SSO_URL)) {
            LOG.info("browser initiated single log out");
            logout(request);
            String callback = request.getParameter("callback");
            String jsonp = String.format("%s({\"status\":\"success\"})", callback);
            response.getWriter().print(jsonp);
            response.getWriter().flush();
            return;
        }

        if (context.getLoginConfig() != null) {
            String authMethod = context.getLoginConfig().getAuthMethod();
            /*
             * Without registering authenticator and its auth method in the container, 
             * it won't even invoke the valve if unknown auth method encountered in app's web.xml.
             * If login-config is omitted then by default auth method is NONE.
             */
            if (AUTH_METHOD.equals(authMethod) || "NONE".equals(authMethod)) {
                super.invoke(request, response);
                return;
            }
        }

        getNext().invoke(request, response);
    }

    @Override
    public boolean authenticate(Request request, HttpServletResponse response, LoginConfig config) throws IOException {
        LOG.info("authenticate()");

        // already authenticated?
        if (request.getUserPrincipal() != null) {
            return true;
        }

        // response from IDP?
        if (request.getParameter("SAMLResponse") != null) {
            return processResponse(request, response);
        }

        // redirect to IDP
        redirectToIdp(request, response);

        return false;
    }

    boolean processResponse(Request request, HttpServletResponse response) throws IOException {
        try {
            Session session = request.getSessionInternal(false);
            if (session == null) {
                throw new Exception("Could not restore original request, no session found");
            }

            String expectedInResponseTo = (String) session.getSession().getAttribute(ID);
            String expectedAudience = (String) session.getSession().getAttribute(AUDIENCE);
            CloudsealPrincipal cloudsealPrincipal = responseValidator.validateResponse(idpXmlParser.getIdpPublicKey(), request.getParameter("SAMLResponse"), expectedInResponseTo, expectedAudience);

            restoreRequest(request, session);

            List<String> roles = new ArrayList<String>(cloudsealPrincipal.getRoles());
            GenericPrincipal principal = new GenericPrincipal(cloudsealPrincipal.getUsername(), null, roles, cloudsealPrincipal);
            // authenticate and save to session (this will create new session)
            register(request, response, principal, getAuthMethod(), principal.getName(), principal.getPassword());

            return true;
        } catch (Exception e) {
            LOG.error(e.getMessage(), e);
            sendForbiddenError(response, e.getMessage());
            return false;
        }
    }

    void redirectToIdp(Request request, HttpServletResponse response) throws IOException {
        try {
            String spIssuer = buildSpIssuer(request);
            String id = idGenerator.generateIdentifier();
            String acsUrl = buildAcsUrl(request);
            request.getSession().setAttribute(ID, id);
            request.getSession().setAttribute(AUDIENCE, spIssuer);
            String samlAuthRequest = authRequestBuilder.generateSamlAuthRequest(spIssuer, acsUrl, spIssuer, id);
            String redirectUrl = idpXmlParser.getSsoUrl() + "?" + samlAuthRequest;
            response.sendRedirect(redirectUrl);
        } catch (Exception e) {
            LOG.error(e.getMessage(), e);
            sendForbiddenError(response, e.getMessage());
        }
    }

    byte[] getResource(String filename) throws IOException {
        ServletContext servletContext = this.context.getServletContext();
        InputStream is;
        if (filename.startsWith("classpath:")) {
            String path = filename.split(":")[1];
            is = this.getClass().getClassLoader().getResourceAsStream(path);
        } else {
            is = servletContext.getResource(filename).openStream();
        }
        try {
            return IOUtils.toByteArray(is);
        } finally {
            IOUtils.closeQuietly(is);
        }
    }

    String buildSpIssuer(Request request) {
        String host = request.getHeader("Host");
        String protocol = "http://";
        if (request.isSecure()) {
            protocol = "https://";
        }
        if (host != null && host.length() > 0) {
            return protocol + host + "/saml/sp";
        } else {
            return protocol + "www.mycorp.com" + "/saml/sp";
        }
    }

    String buildAcsUrl(Request request) {
        return request.getRequestURL().toString();
    }

    void sendForbiddenError(HttpServletResponse response, String error) throws IOException {
        response.sendError(HttpServletResponse.SC_FORBIDDEN, error);
    }

    public void setIdp(String idp) {
        this.idp = idp;
    }

    public void setKeystore(String keystore) {
        this.keystore = keystore;
    }

    public void setKeystorePassword(String keystorePassword) {
        this.keystorePassword = keystorePassword;
    }

    public void setKey(String key) {
        this.key = key;
    }

    public void setKeyPassword(String keyPassword) {
        this.keyPassword = keyPassword;
    }

    public void setAppId(String appId) {
        this.appId = appId;
    }

}
